/* whyConserved - Try and analyse why a particular thing is conserved. */
#include "common.h"
#include "linefile.h"
#include "hash.h"
#include "cheapcgi.h"
#include "hdb.h"
#include "jksql.h"
#include "bits.h"
#include "featureBits.h"

void usage()
/* Explain usage and exit. */
{
errAbort(
  "whyConserved - Try and analyse why a particular thing is conserved\n"
  "usage:\n"
  "   whyConserved database chromosome homologyTrack\n"
  "options:\n"
  "   -xxx=XXX\n"
  );
}


void explainSome(Bits *homo, Bits *once, Bits *bits, char *chrom, int chromSize, 
	struct sqlConnection *conn, char *trackSpec, char *homologyTrack)
/* Explain some of homology. */
{
int trackSize, homoSize, andSize, cumSize, newSize;

homoSize = bitCountRange(homo, 0, chromSize);
bitClear(bits, chromSize);
if (trackSpec != NULL)
    {
    fbOrTableBits(bits, trackSpec, chrom, chromSize, conn);
    trackSize = bitCountRange(bits, 0, chromSize);
    bitAnd(bits, homo, chromSize);
    andSize = bitCountRange(bits, 0, chromSize);
    bitAnd(bits, once, chromSize);
    newSize = bitCountRange(bits, 0, chromSize);
    bitNot(bits, chromSize);
    bitAnd(once, bits, chromSize);
    cumSize = homoSize - bitCountRange(once, 0, chromSize);
    }
else
    {
    trackSpec = homologyTrack;
    trackSize = andSize = homoSize;
    cumSize = newSize = 0;
    }

printf("%-21s %8d %8d %5.2f%% %6.2f%% %5.2f%% %5.2f%%\n",
	trackSpec, trackSize, andSize, 
	100.0*trackSize/chromSize,
	100.0*andSize/homoSize,
	100.0*newSize/homoSize,
	100.0*cumSize/homoSize);
}

void whyConserved(char *database, char *chrom, char *homologyTrack)
/* whyConserved - Try and analyse why a particular thing is conserved. */
{
struct sqlConnection *conn;
int chromSize;
Bits *homo = NULL;
Bits *bits = NULL;
Bits *once = NULL;

hSetDb(database);
conn = hAllocConn();
chromSize = hChromSize(chrom);

homo = bitAlloc(chromSize);
bits = bitAlloc(chromSize);
once = bitAlloc(chromSize);

/* Get homology bitmap and set once mask to be the same. */
fbOrTableBits(homo, homologyTrack, chrom, chromSize, conn);
bitOr(once, homo, chromSize);

/* printHeader */
printf("%-21s %8s %8s %5s %6s  %5s  %5s \n",
	"Track Specification", "track", "overlap", 
	"track", "ovlp",  "new",  "cum");
printf("%-21s %8s %8s %5s %6s  %5s  %5s \n",
	"", "size", "size", 
	" % ", " % ",  " % ",  " % ");
printf("---------------------------------------------------------------------\n");

/* Whittle awway at homology... */
explainSome(homo, once, bits, chrom, chromSize, conn, NULL, homologyTrack);
// explainSome(homo, once, bits, chrom, chromSize, conn, "simpleRepeat", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "rmsk", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "sanger22:CDS:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "refGene:CDS:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "sanger22:exon:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "refGene:exon:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "rnaGene", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "mrna:exon:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "intronEst:exon:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "xenoMrna:exon:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "xenoEst:exon:10", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "sanger22:promoter:200", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "refGene:promoter:200", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "mrna:promoter:200", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "est", NULL);
explainSome(homo, once, bits, chrom, chromSize, conn, "ensGene", NULL);
hFreeConn(&conn);
}


void testBits()
/* Test bitmap ops. */
{
Bits *a = bitAlloc(100), *b = bitAlloc(100), *c = bitAlloc(100);

bitSetRange(a, 10, 10);
bitSetRange(a, 30, 10);
bitSetRange(b, 30, 10);
bitSetRange(b, 50, 10);
bitSetRange(c, 10, 10);
bitSetRange(c, 30, 10);

uglyf("a = %d\n", bitCountRange(a, 0, 100));
uglyf("b = %d\n", bitCountRange(b, 0, 100));
bitAnd(a, b, 100);
uglyf("a&b = %d\n", bitCountRange(a, 0, 100));
bitOr(c, b, 100);
uglyf("a|b = %d\n", bitCountRange(c, 0, 100));
bitNot(c, 100);
uglyf("!(a|b) = %d\n", bitCountRange(c, 0, 100));
}

int main(int argc, char *argv[])
/* Process command line. */
{
cgiSpoof(&argc, argv);
if (argc != 4)
    usage();
whyConserved(argv[1], argv[2], argv[3]);
return 0;
}
